/*
 * Copyright (c) 2016-2019 Deephaven Data Labs and Patent Pending
 */

package com.illumon.demo.sample;

import com.fishlib.configuration.Configuration;
import com.fishlib.io.log.LogLevel;
import com.fishlib.io.logger.Log4jAdapter;
import com.fishlib.io.logger.Logger;
import com.fishlib.logger.ProcessUnsafeLogCreator;
import com.fishlib.net.impl.nio.NIODriver;
import com.fishlib.util.process.ProcessEnvironment;
import com.illumon.iris.db.tables.TableDefinition;
import com.illumon.iris.db.tables.databases.Database;
import com.illumon.iris.db.tables.remote.RemoteDatabase;
import com.illumon.iris.db.tables.remotequery.ContextAwareRemoteQuery;
import com.illumon.iris.db.tables.remotequery.RemoteProcessingRequest;
import com.illumon.iris.db.tables.remotequery.RemoteQueryClient;
import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.io.UncheckedIOException;

/**
 * Simple example for executing a Select MAX() query against an Iris intraday table and partition.
 *
 * This example creates an Iris database connection (RemoteDatabase), and also creates an inner
 * query class to allow the query steps to be executed as a single request to the query server.
 * The query class extends ContextAwareRemoteQuery, and an instance of this class is then passed
 * to the RemoteDatabase executeQuery method to execute the contained query and return the result.
 *
 * For a simpler example that uses ad-hoc method calls against the RemoteDatabase object to execute
 * the same query, see SimpleMaxValue.java.
 *
 * This class requires a properties file that includes RemoteQueryDispatcherParameters.host, to tell
 * it what query host to connect to. It also requires -DAuthenticationClientManager.defaultPrivateKeyFile
 * passed as one of the JVM arguments, to provide a private key to use for authentication to Iris.
 *
 * Setting up a key pair for authentication:
 * 1. Set up Iris Authorization keys on the Iris server by executing the following on the server:
 * /usr/illumon/latest/bin/generate_iris_keys <irisUserName>
 * cat pub-<irisUserName>.base64.txt >> /etc/sysconfig/illumon.d/resources/dsakeys.txt
 *
 * 2. Copy the user's private auth key ( priv-<irisUserName>.base64.txt ) to the user's home directory,
 * or somewhere else that is only accessible to the user.
 */

public class SimpleMaxValueCARQ {
    private static final String PROCESS_NAME = "SimpleMaxValueDemoCARQ";

    private static void printUsageAndExit(@NotNull final String message) {
        System.err.println(message);
        System.err.println("Usage: SimpleMaxValueCARQ namespace table_name column_name [partitionValue]");
        System.exit(1);
    }

    private static void prepareEnvironment() {
        // the Configuration object reads the Iris property file, which has named key-value pairs that control the
        // operation of the Iris system.
        final Configuration configuration = Configuration.getInstance();

        // the log object is used by Iris components, and produces output in a PROCESS_NAME.log.<date> with
        // a hard link to PROCESS_NAME.log.current (e.g., DemoIntegration.log.current).  The log files are stored
        // in <workspace>/../logs
        Logger log = new ProcessUnsafeLogCreator(PROCESS_NAME, configuration).createLogger(PROCESS_NAME, LogLevel.INFO);

        // All log4j output should be sent to the Logger we created on the line above
        Log4jAdapter.sendLog4jToLogger(log);

        // The Iris system uses a wrapper around Java NIO for most network communication and job scheduling, this
        // initializes the communications subsystem with our log
        NIODriver.init(log);

        // The ProcessEnvironment provides hooks for consistent startup and shutdown initialization across Iris processes.
        ProcessEnvironment.basicInteractiveProcessInitialization(configuration, PROCESS_NAME, log);
    }

    /**
     * This static inner class extends ContextAwareRemoteQuery to provide a query class that is specifically
     * designed to execute the Select MAX() operation against a specified table, partition, and column.
     *
     * This class must also be available to the query worker, so the worker will be able to resolve it when
     * executeQuery is called. This can be done by deploying this class in a jar to the query server
     * or by pushing it to the worker. This example pushes it to the worker. See the rqc.pushClasses
     * call, below.
     */
    public static class GetMaxValue extends ContextAwareRemoteQuery<Object> {
        private final String namespace;
        private final String table;
        private final String partitionValue;
        private final String partitionColumn;
        private final String column;

        /**
         * Classes that extend ContextAwareRemoteQuery must override execute. This
         * is where the specific query goes that we want to execute. Note that only
         * the Database object is passed in here (it is passed in by the executeQuery
         * method called against the RemoteDatabase). Other parameters must be defined
         * as class variables that are set by the class constructor.
         */
        @Override
        public Object execute(Database db) {
            return db.i(namespace, table).where(partitionColumn + "=" + partitionValue).select(column).maxBy().getColumn(column).get(0);
        }

        /**
         * The class constructor, which takes the parameters needed for this query;
         * these parameter values will be used when execute() is called.
         * @param namespace Namespace in which to find the table.
         * @param tableName Name of the table against which the query will be executed.
         * @param partitionColumn Partition column to be used when constructing the where clause in the query.
         * @param partitionValue Value to be used in conjunction with the partitionColumn in th where clause.
         * @param column The column for which the maximum value will be returned.
         */
        public GetMaxValue(@NotNull String namespace, @NotNull String tableName, @NotNull String partitionColumn, @NotNull String partitionValue, @NotNull String column) {
            this.namespace = namespace;
            this.table = tableName;
            this.column = column;
            this.partitionColumn = partitionColumn;
            this.partitionValue = partitionValue;
        }
    }

    /**
     * Run a sample query to get the max value from a specified column within an optionally specified partition on
     * an <b>intraday</b> table
     *
     * @param args 1: The namespace to look up.
     *             2: The name of the table to look up.
     *             3: The column to look up (must be amenable to a 'max' operation).
     *             4: (Optional) The partition value to use. If not specified, will use currentDateNy.
     */
    public static void main(@NotNull final String... args) {
        if (args.length < 3) {
            printUsageAndExit("Invalid number of arguments");
            return;
        }

        // Parse parameter values from the command line
        final String namespace = args[0];
        final String table = args[1];
        final String column = args[2];
        final String partitionValue;
        if (args.length > 3) {
            partitionValue = "`" + args[3] + "`";
        } else {
            partitionValue = "currentDateNy()";
        }

        final long overallStartNanos = System.nanoTime();

        // Set up logging and process management
        prepareEnvironment();

        // Create an Iris client
        final RemoteQueryClient rqc;
        try {
            rqc = new RemoteQueryClient();
        } catch (IOException e) {
            throw new UncheckedIOException("Failed to create RemoteQueryClient", e);
        }

        /*
         * Attempt to push the GetMaxValue class to the worker. This class will be needed by
         * by the worker when executeQuery() is called.
         */
        try {
            rqc.pushClasses(GetMaxValue.class);
        } catch (IOException e) {
            throw new UncheckedIOException("Failed to push GetMaxValue class", e);
        }

        // Create a connection to the remote query server specified by RemoteQueryDispatcherParameters.host
        final RemoteDatabase db;
        try {
            db = rqc.getRemoteDB(RemoteProcessingRequest.DEFAULT_MAX_HEAP_SIZE_MB, PROCESS_NAME, RemoteProcessingRequest.DEFAULT_WORKER_DEBUG_ENABLED);
        } catch (IOException e) {
            throw new UncheckedIOException("Failed to create RemoteDatabase", e);
        }

        // Retrieve the partition column name from the TableDefinition for this table
        final TableDefinition tableDefinition = db.getTableDefinition(namespace, table);
        if(tableDefinition == null) {
            throw new IllegalArgumentException("Table " + table + " doesn't exist in namespace " + namespace);
        }
        final String partitionColumn = tableDefinition.getPartitioningColumns().get(0).getName();

        final long startNanos = System.nanoTime();

        /*
         * Attempt to call executeQuery. Within the call, instantiate a new GetMaxValue query object and pass in
         * the parameters that will be needed to execute the query. The executeQuery method will use the
         * GetMaxValue class that was pushed through the RemoteQueryClient session to interpret the query object
         * and invoke its execute() method. The execute() method will execute the query and return the Max value.
         */
        final Object value;
        try {
            value = db.executeQuery(new GetMaxValue(namespace, table, partitionColumn, partitionValue, column));
        } catch (IOException e) {
            throw new UncheckedIOException("Failed to execute remote query.", e);
        }

        final long endNanos = System.nanoTime();

        // Output the results
        System.out.println();
        System.out.println("Test complete for Max(" + column + ") for " + namespace + "/" + table + " for partition value of " + partitionValue.replace("`",""));
        System.out.println("Query duration " + (endNanos - startNanos) + " ns");
        System.out.println("Overall duration " + (endNanos - overallStartNanos) + " ns");
        System.out.println();
        if (value != null) {
            System.out.println("Result: " + value.toString());
        } else {
            System.out.println("(No intraday data found for that namespace and table on the specified date)");
        }
        System.out.println();
    }
}
